package nl.dancingbear.tunit;

import java.io.IOException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.concurrent.TimeoutException;

import junit.framework.TestCase;
import aterm.ATermFactory;
import aterm.pure.PureFactory;

/**
 * The TUnitTestCase class is a base class for testing toolbus scripts.
 * This class can start / stop the toolbus and connect / disconnect
 * stubs from the toolbus.
 *
 * @author Bas Basten & Anton Lycklama a Nijeholt
 * @date 23-02-2007
 */
public abstract class TUnitTestCase extends TestCase {
    private ArrayList<ToolStub> stubs = new ArrayList<ToolStub>();
    private Process toolbusProcess;
    private int toolbusPort;

    protected ATermFactory factory = new PureFactory();

    /**
     * Add stubs to the test case.
     *
     * @param toolName The name for the tool stub.
     * @author Bas Basten
     * @author Anton Lycklama a Nijeholt
     * @date 23-02-2007
     */
    protected final ToolStub createToolStub(String toolName) {
        ToolStub toolStub = new ToolStub(this.factory, toolName);
        this.stubs.add(toolStub);
        return toolStub;
    }

    /**
     * Get the absolute top path of the current package.
     *
     * @author Bas Basten
     * @author Anton Lycklama a Nijeholt
     * @date 06-03-2007
     */
    public String getTopSrcDir() {
        // Get the environment variable BUILDFILE that is passed
        // from the ant file that is generated from the build scripts.
        String topSrcDir = System.getenv("BUILDFILE");

        if (null != topSrcDir){
            int pos = topSrcDir.indexOf("/meta-build.ant");
            if (pos != -1) {
                return topSrcDir.substring(0, pos);
            }
        }

        // When the unit test is not called from an ant script, we just
        // return the current working directory. (For use in Eclipse)
        return ".";
    }

    /**
     * Starts the toolbus at a given port and executes a script. If a script
     * uses include scripts the searchPath will be used to include the include
     * scripts from.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @date 23-02-2007
     * @param searchPath The directories to search for the included script files of the specified script.
     * @param scriptName The script to be executed.
     * @param toolbusPort The port for the toolbus to listen to.
     *
     */
    protected void startToolbus(String searchPath, String scriptName, int toolbusPort) throws IOException, TimeoutException {
        this.toolbusPort = toolbusPort;

        final long timeout = 10000;
        long startTime = System.currentTimeMillis();

        // Launch the toolbus with script init.tb
        toolbusProcess = Runtime.getRuntime().exec("toolbus -TB_PORT " + String.valueOf(toolbusPort) + " -I" + searchPath + " " + scriptName);

        // Wait until the toolbus is initialized and started.
        // Listen to the error stream of the toolbus, the INIT process
        // will tell us when the toolbus is ready.
        StringBuilder builder = new StringBuilder();
        String toolbusStarted = new String("Toolbus started...\n");

        while (!builder.toString().contains(toolbusStarted)) {
            byte[] buffer = new byte[100];
            int bytesRead = this.toolbusProcess.getErrorStream().read(buffer);
            if (bytesRead > 0) {
                String toolbusOutput = new String(buffer, 0, bytesRead);
                builder.append(toolbusOutput, 0, bytesRead);
                System.out.print(toolbusOutput);
            }

            if (System.currentTimeMillis() >= (startTime + timeout)) {
                throw new TimeoutException("");
            }
        }

        StreamRedirector redirector = new StreamRedirector(this.toolbusProcess.getErrorStream(), System.out);
        Thread redirectorThread = new Thread(redirector);
        redirectorThread.start();
    }

    /**
     * Stops the toolbus.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @date 23-02-2007
     */
    protected void stopToolbus() {
        if (this.toolbusProcess != null) {
            this.toolbusProcess.destroy();
        }
    }

    /**
     * Connects the tool stubs to the toolbus.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @date 23-02-2007
     */
    protected void connectToolStubs() throws UnknownHostException, IOException {
        String[] toolbusArgs = { "-TB_PORT", String.valueOf(this.toolbusPort) };

        for (ToolStub toolStub : this.stubs) {
            toolStub.init(toolbusArgs);
            toolStub.connect();

            Thread thread = new Thread(toolStub);
            thread.start();
        }
    }

    /**
     * Disconnects the tool stubs from the toolbus.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @date 23-02-2007
     */
    protected void disconnectToolStubs() {
        for (ToolStub toolStub : this.stubs) {
            toolStub.disconnect();
        }
    }
}
